#
#  ddt.pl - perl functions for the DDT commands
#
#  Copyright (C) 1992, 1993 PUUG - Grupo Portugues de Utilizadores
#                                  do Sistema UNIX
#                1992, 1993 FCCN - Fundacao para o Desenvolvimento dos Meios 
#                                  Nacionais de Calculo Cientifico 
#
#  Authors: Jorge Frazao de Oliveira <frazao@puug.pt>
#           Artur Romao <artur@dns.pt>
#
#  This file is part of the DDT package, Version 2.0.
#
#  Permission to use, copy, modify, and distribute this software and its 
#  documentation for any purpose and without any fee is hereby granted, 
#  provided that the above copyright notice appear in all copies.  Neither 
#  PUUG nor FCCN make any representations about the suitability of this
#  software for any purpose.  It is provided "as is" without express or 
#  implied warranty.


#
# canonicalises to lowercase
#
sub tolower {
        local($s) = @_;

        $s =~ tr/A-Z/a-z/;
        return($s);
}


#
# canonicalises to uppercase
#
sub toupper {
        local($s) = @_;

        $s =~ tr/a-z/A-Z/;
        return($s);
}


#
# tests to see if a given $elem is in a $set
# $set is a $sep separated list (string) of elements
#
sub in {
	local($elem, $set, $sep) = @_;

	return ($sep . $set . $sep) =~ ($sep . $elem . $sep);
}

#
# tests to see if $name is "inside" $zone
# if $zone is of the form "a.b.c", compares $name with 
# the regexp /[.]a[.]b[.]c$/
#
sub is_below_cut {
        local($zone, $name) = @_;

        $zone =~ s/"[.]"/[.]/g;

        return $name =~ ("[.]" . $zone . "\$");
}


#
# $name is inside a (strict) sub-zone of $zone
#
sub in_sub_zone {
    	local($name, $zone) = @_;
	local(@NSlist, $doamin, $server);

    	@NSlist = split($LISTsep, $NS{$zone});

	while (@NSlist) {
		($domain, $server) = split($ELEMsep1, shift(@NSlist));

		if ($domain ne $zone && &is_below_cut($domain, $name)) {
			return 1;
	    	}
	}

    	return 0;
}


#
# $name belongs to this zone (not one of its sub-zones)
#
sub in_THIS_zone {
    	local($name, $zone) = @_;

    	return &is_below_cut($zone, $name) && !&in_sub_zone($name, $zone);
}


#
# a name was found; build appropriately, cannonicalised
# to lower case and expanded with origin, if necessary
#
sub make_name {
        local($word, $origin) = @_;
     
	if (substr($word, length($word)) ne ".") {
		if ($origin eq "") {
			return &tolower($word);
		}
		elsif ($origin eq ".") {
			return &tolower($word) . $origin;
		}
		else {
			return &tolower($word) . "." . &tolower($origin);
			}
	}
}


#
# appends $elem to the end of $list (a string)
# the elements of $list are separated by commas (",")
#
sub add_list {
	local($list, $elem) = @_;

    	if ($list ne "") {
		return $list . "," . $elem;
	}
	else {
		return $elem;
	}

}


$[     = 1;		# set array base to 1
$,     = ' ';      	# set output field separator
$\     = "\n";        	# set output record separator

# set the number of left padding spaces, according to the message level
@Lpad  = ("  ", "    ", "      ", "        ");

# set the overall message level, given as an argument, LEVEL as default
# =()<$Level = ($#ARGV == 2 && $ARGV[1] eq "-l") ? $ARGV[2] : @<LEVEL>@;>()=
$Level = ($#ARGV == 2 && $ARGV[1] eq "-l") ? $ARGV[2] : 4;

$LISTsep  = ",";
$ELEMsep1 = ":";
$ELEMsep2 = ";";

# set initial values
$Origin      = "";
$Name        = ".";
$Zone	     = "";


1;
