.\"
.\" MAN PAGE COMMENTS to
.\"
.\"     Jorge Frazao de Oliveira
.\"     Portuguese Unix User's Group
.\"     frazao@puug.pt
.\"
.\" Copyright (C) 1992 Portuguese Unix User's Group
.\" Everyone is permitted to copy and distribute verbatim copies
.\" of this license document, but changing it is not allowed.
.\"
.\" Distributed with 'DDT' version 1.0.
.\"
.TH ddt 1 "May 6 1992" "" "Domain Debug Tools Manual"
.SH NAME
ddt \- domain debug tools
.SH DESCRIPTION
The DNS database is a distributed database where almost all the distributed
bindings of the TCP/IP application level take place. 
This database is managed by several primary and secondary servers
running in a master/slave fashion.
During updates, there are temporary inconsistencies that are more or less
tolerated by the different user services.
By its nature, the database is managed in a descentralized way, with hundreds,
or even thousands, of individuals envolved. 
This is a very interesting and valuable feature of this application.
However, in conjunction with the fact that inconsistencies must be tolerated,
this collective managing style may lead to the introduction of
hard to find errors.
Sometimes, some of 
these errors can introduce several problems and resource usage far from ideal.
.PP
.I Ddt
(domain debug tools) is a package to help zone administrators to avoid
as much as possible these problems.
The package is a set of commands that allow administrators to analyze any
portion of the DNS tree.
.PP
.I Ddt
works on cached data files because, as such, it can be reasonable efficient.
A command is available to try to cache the "real world" situation
as close as possible.
This command,
.I ddt-xfer(1),
is a slightly modified version of the bind
.I named-xfer(8)
that allows the transfer to the cache of a zone and, optionally, 
all its descendents zones. If you have enough disk and bandwith,
you can cache all the zones of the world with one only command.
Probably, at end, you should start again because some of
the cached zones will not reflect the real ones anymore.
.PP
All the available commands can receive one or more zone files as parameters.
So, one command can spread its analysis throughout a set of zones.
With the exception of 
.I ddt-xfer(1),
all commands are written in gawk (GNU awk).
.PP
The commands available are:
.RS 2
.IP "\fIddt-xfer\fP" 10
to cache one or more zones.
.IP "\fIsoac\fP"
to analyze the RRs that describe the top node of the zone, i.e., 
the NS RRs that list the servers for the zone and a single SOA RR
that describe zone management parameters.
.IP "\fIrrc\fP"
to check the semantic of the RRs, like valid host names, names without
trailing dot, etc.
.IP "\fIgrc\fP"
to analyze the glue records for a zone.
.IP "\fImxc\fP"
to analyze the MX RRs defined within a zone.
.IP "\fIrmc\fP"
to analyze the reverse mapping.
.RE
.sp
.PP
Besides the commands that allow the analysis of the zone, there are
some small commands to obtain several accounts, like
number of hosts, number of nets, number of domains, most popular names, etc.
In conjunction with some user written awk scripts, it is easy to obtain
a lot of other information.
These commands are the following filters:
.RS 2
.IP "\fIexpand\fP" 13
replaces the owner name of the RR by an absolute or fully qualified name.
.IP "\fIhosts\fP"
returns all hosts in the zone.
.IP "\fIhosts-addr\fP"
returns all hosts with their ip-addresses.
.IP "\fIhosts-domain\fP"
returns all domains defined in the zones with the number of hosts per domain.
.IP "\fIcnames\fP"
returns all nicknames in the zone.
.IP "\fInets\fP"
returns all nets in the zone with the number of hosts per network.
.IP "\fInames-stat\fP
does some statistics about names.
.RE
.sp
.PP
To make the interpretation of the several messages returned by the
available commands easier, 
those messages are divided into four levels of severity:
.RS 2
.IP "[Warning 1] You should investigate it now!"
.IP "[Warning 2] Don't forget this warning!"
.IP "[Warning 3] You should analyze it someday!"
.IP "[Comment] You can ignore it / "
.RS 10
DDT has not enough data to take an effective decision.
.RE
.RE
.PP
Messages belonging to [Warning 1] level are the most serious. On the other
hand, messages belonging to [Comment] level are less important, but
this doesn't mean that these messages can be completely ignored.
Unfortunately,  there are some serious
errors that were included in the this level because, when the message
is displayed, DDT has not enough
data (zone files) to take an effective decision.
.PP
The DDT_LEVEL environment variable is used to select what messages
should be displayed. If it has the value 1, only messages
belonging to [Warning 1] level are displayed. If it is 2 are displayed all 
messages belonging to [Warning 1] and [Warning 2] levels and so on.
.sp
.PP
The commands 
.I ddt-xfer(1),
.I grc(1),
.I mxc(1),
.I rrc(1) 
.I rmc(1) 
and
.I saoc (1)
are described in separate manual pages. 
The commands for gathering statistics about zones are described below 
with some examples.
.PP
.IP 1) 3
Get all zones below the domain fr. (inclusive):
.nf

.RS 2
ddt-xfer -z fr -f . -r inria.inria.fr layon.inria.fr
.RE
.fi
.sp
.PP
.IP 2) 3
Get the zone ul.pt only if the serial number had changed:
.nf

.RS 2
ddt-xfer -z ul.pt -f . -s 199202140 muttley.ul.pt dec4pt.puug.pt
.RE
.fi
.sp
.PP
.IP 3) 3
Generate the portuguese statistics of hostnames 
(All zones below pt. are in /usr/local/ddt.cache/pt directory):
.nf

.RS 2
expand /usr/local/ddt.cache/pt/* | hosts | names-stat | sort -rn
.RE
.fi
.sp
.PP
.IP 4) 3
Count hosts in a zone beginning at the node inesc.pt 
(The zone inesc.pt is in the file /usr/local/ddt.cache/pt/inesc.pt):
.nf

.RS 2
expand /usr/local/ddt.cache/pt/inesc.pt | hosts | wc -l
.RE
.fi
.sp
.PP
.IP 5) 3 
Print the number of hosts by nets in Italy 
(All zones below it. are in /usr/local/ddt.cache/it directory):
.nf

.RS 2
expand /usr/local/ddt.cache/it/* | hosts-addr | nets | sort -rn
.RE
.fi
.sp
.PP
Comments and suggestions are welcome.
.SH "SEE ALSO"
ddt-xfer(1), soac(1), rrc(1), mxc(1), grc(1), rmc(1)
.SH AUTHOR
Jorge Fraza\*~o

